/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.xta;

import java.io.IOException;
import java.util.List;

import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;
import org.apache.commons.lang3.StringUtils;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;

import de.ozgcloud.xta.client.XtaClient;
import de.ozgcloud.xta.client.XtaClientFactory;
import de.ozgcloud.xta.client.config.XtaClientConfig;
import de.ozgcloud.xta.client.exception.XtaClientInitializationException;
import de.ozgcloud.xta.client.model.XtaIdentifier;

@Log4j2
@Configuration
@RequiredArgsConstructor
public class XtaServiceConfiguration {

	static final String URI_TEMPLATE = "https://%s/MB_XTA-WS/XTA210";

	static final String XDOMEA_0201_MESSAGE_TYPE = "Geschaeftsgang.Geschaeftsgang.0201";

	static final String FIM_MESSAGE_TYPE_PREFIX = "fim.";

	private final XtaProperties properties;

	@Bean
	XtaClient xtaClient() throws XtaClientInitializationException, IOException {
		XtaClientConfig xtaClientConfig = createClientConfig();
		XtaClientFactory clientFactory = XtaClientFactory.from(xtaClientConfig);
		return clientFactory.create();
	}

	XtaClientConfig createClientConfig() throws IOException {

		var clientCertKeyStore = XtaClientConfig.KeyStore.builder()
				.content(properties.getKeyStore().getFile().getContentAsByteArray())
				.type(properties.getKeyStore().getType())
				.password(properties.getKeyStore().getPassword())
				.build();

		return XtaClientConfig.builder()
				.clientIdentifiers(mapIdentifiers(properties.getIdentifiers()))
				.managementServiceUrl(getManagementPortUrl())
				.sendServiceUrl(getSendPortUrl())
				.msgBoxServiceUrl(getMsgBoxPortUrl())
				.clientCertKeystore(clientCertKeyStore)
				.isMessageSupported(this::isSupportedMessageType)
				.build();
	}

	private List<XtaIdentifier> mapIdentifiers(List<String> identifiers) {
		return identifiers.stream()
				.map(identifier -> XtaIdentifier.builder().value(identifier).build())
				.toList();
	}

	String getBaseUrl() {
		return URI_TEMPLATE.formatted(properties.getServer().getAddress());
	}

	private String getMsgBoxPortUrl() {
		return getBaseUrl() + "msgBoxPort.svc";
	}

	private String getManagementPortUrl() {
		return getBaseUrl() + "managementPort.svc";
	}

	private String getSendPortUrl() {
		return getBaseUrl() + "sendPort.svc";
	}

	boolean isSupportedMessageType(de.ozgcloud.xta.client.model.XtaMessageMetaData metaData) {
		if (isXDomeaMessageType(metaData.messageTypeCode()) || isFimMessageType(metaData.messageTypeCode())) {
			return true;
		}
		LOG.warn("Ignoring XTA-Message of type '{}'.", metaData.messageTypeCode());
		return false;
	}

	private boolean isXDomeaMessageType(String messageType) {
		return StringUtils.equals(messageType, XDOMEA_0201_MESSAGE_TYPE);
	}

	private boolean isFimMessageType(String messageType) {
		return StringUtils.startsWith(messageType, FIM_MESSAGE_TYPE_PREFIX);
	}

}
