/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.xdomea;

import static de.ozgcloud.eingang.xdomea.XdomeaMessageDataMapper.*;
import static de.ozgcloud.eingang.xdomea.XdomeaMessageDataTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Collection;
import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;

class XdomeaMessageDataMapperTest {

	@Spy
	@InjectMocks
	private XdomeaMessageDataMapper fileClassifier;

	@Mock
	private XdomeaXMLValueReader valueReader;

	@DisplayName("map incoming files to representations attachments pair")
	@Nested
	class TestMapIncomingFilesToRepresentationsAttachmentsPair {

		@Mock
		private List<IncomingFile> incomingFileList;

		@DisplayName("should map representations")
		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		void shouldMapRepresentations(boolean useUuidPrefix) {
			mockXdomeaMessageData(useUuidPrefix);

			var pair = fileClassifier.mapIncomingFilesToRepresentationsAttachmentsPair(incomingFileList);

			var representationFilenames = pair.representations().stream().map(IncomingFile::getName).toList();
			assertThat(representationFilenames).containsExactly(XDOMEA_XML_NAME, REPR_XML_NAME, REPR_PDF_NAME);
		}

		@DisplayName("should map attachments")
		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		void shouldMapAttachments(boolean useUuidPrefix) {
			mockXdomeaMessageData(useUuidPrefix);

			var pair = fileClassifier.mapIncomingFilesToRepresentationsAttachmentsPair(incomingFileList);

			var attachmentFilenames = pair.attachments().stream()
					.map(IncomingFileGroup::getFiles)
					.flatMap(Collection::stream)
					.map(IncomingFile::getName)
					.toList();
			assertThat(attachmentFilenames).containsExactly(ATTATCHMENT_XML_NAME, ATTATCHMENT_PNG_NAME, ATTATCHMENT_PDF_NAME);
		}

		private void mockXdomeaMessageData(boolean useUuidPrefix) {
			doReturn(XdomeaMessageDataTestFactory.create(useUuidPrefix)).when(fileClassifier).mapIncomingFilesToXdomeaMessageData(incomingFileList);
		}

	}

	@DisplayName("find Xdomea XML file")
	@Nested
	class TestFindXdomeaXmlFile {

		private static final String FILE_NAME_WITHOUT_SUFFIX = "some-file-name.xml";
		private static final String FILE_NAME_WITH_SUFFIX = "some-file-name" + XDOMEA_0201_XML_FILE_NAME_SUFFIX;

		@DisplayName("should throw if not found")
		@Test
		void shouldThrowIfNotFound() {
			var incomingFilesWithout = List.of(IncomingFileTestFactory.createBuilder().name(FILE_NAME_WITHOUT_SUFFIX).build());

			assertThatThrownBy(() -> fileClassifier.findXdomeaXMLFile(incomingFilesWithout))
					.isInstanceOf(TechnicalException.class);
		}

		@DisplayName("should return if found")
		@Test
		void shouldReturnIfFound() {
			var targetIncomingFile = IncomingFileTestFactory.createBuilder().name(FILE_NAME_WITH_SUFFIX).build();
			var incomingFilesWith = List.of(
					IncomingFileTestFactory.createBuilder().name(FILE_NAME_WITHOUT_SUFFIX).build(),
					targetIncomingFile
			);

			var primaryRepresentation = fileClassifier.findXdomeaXMLFile(incomingFilesWith);

			assertThat(primaryRepresentation).isEqualTo(targetIncomingFile);
		}
	}

	@DisplayName("find primary representation name")
	@Nested
	class TestFindPrimaryRepresentationName {
		private static final String FILE_NAME_WITHOUT_XML_SUFFIX = "some-file-name.pdf";
		private static final String FILE_NAME_WITHOUT_XML_SUFFIX2 = "some-file-name.xml.pdf";
		private static final String FILE_NAME_WITH_XML_SUFFIX = "some-file-name.xml";
		private static final String FILE_NAME_WITH_XML_SUFFIX2 = "some-file-name.pdf.xml";

		@DisplayName("should throw if no xml suffix")
		@Test
		void shouldThrowIfNoXmlSuffix() {
			var listWithoutSuffix = List.of(FILE_NAME_WITHOUT_XML_SUFFIX, FILE_NAME_WITHOUT_XML_SUFFIX2);

			assertThrows(TechnicalException.class, () -> fileClassifier.findPrimaryRepresentationName(listWithoutSuffix));
		}

		@DisplayName("should return first with xml suffix")
		@Test
		void shouldReturnFirstWithXmlSuffix() {
			var listWithSuffix = List.of(FILE_NAME_WITHOUT_XML_SUFFIX, FILE_NAME_WITH_XML_SUFFIX, FILE_NAME_WITHOUT_XML_SUFFIX2,
					FILE_NAME_WITH_XML_SUFFIX2);

			var fileName = fileClassifier.findPrimaryRepresentationName(listWithSuffix);

			assertThat(fileName).isEqualTo(FILE_NAME_WITH_XML_SUFFIX);
		}
	}

	@DisplayName("classify attachments and representations")
	@Nested
	class TestClassifyAttachmentsAndRepresentations {

		@Mock
		private IncomingFile xdomeaXMLFile;

		private List<IncomingFile> incomingFileList;

		@BeforeEach
		void mock() {
			incomingFileList = Stream.of(
							XDOMEA_XML_NAME,
							REPR_XML_NAME,
							REPR_PDF_NAME,
							ATTATCHMENT_XML_NAME,
							ATTATCHMENT_PNG_NAME,
							ATTATCHMENT_PDF_NAME
					)
					.map(name -> IncomingFileTestFactory.createBuilder().name(name).build())
					.toList();
			doReturn(xdomeaXMLFile).when(fileClassifier).findXdomeaXMLFile(incomingFileList);

			when(xdomeaXMLFile.getName()).thenReturn(XDOMEA_XML_NAME);

			var representationFileNames = List.of(REPR_PDF_NAME, REPR_XML_NAME);
			when(valueReader.readRepresentationFileNames(xdomeaXMLFile)).thenReturn(representationFileNames);
			doReturn(REPR_XML_NAME).when(fileClassifier).findPrimaryRepresentationName(representationFileNames);
		}

		@DisplayName("should contain xdomea metadata file")
		@Test
		void shouldContainXdomeaMetadataFile() {
			var classification = doClassify();

			var primaryDocument = classification.metadataFile();
			assertThat(primaryDocument.getName()).isEqualTo(XDOMEA_XML_NAME);
		}

		@DisplayName("should contain representations")
		@Test
		void shouldContainRepresentations() {
			var classification = doClassify();

			var resultRepresentationFileNames = classification.representations().stream()
					.map(IncomingFile::getName)
					.toList();
			// Expect that the primary representation xml file is moved to the first position
			assertThat(resultRepresentationFileNames).isEqualTo(List.of(REPR_XML_NAME, REPR_PDF_NAME));
		}

		@DisplayName("should contain attachments")
		@Test
		void shouldContainAttachments() {
			var classification = doClassify();

			var resultAttachmentFileNames = classification.attachments()
					.getFirst()
					.getFiles().stream()
					.map(IncomingFile::getName)
					.toList();
			assertThat(resultAttachmentFileNames).isEqualTo(List.of(ATTATCHMENT_XML_NAME, ATTATCHMENT_PNG_NAME, ATTATCHMENT_PDF_NAME));
		}

		@DisplayName("should have attachment group name")
		@Test
		void shouldHaveAttachmentGroupName() {
			var classification = doClassify();

			var resultAttachmentGroupName = classification.attachments()
					.getFirst()
					.getName();
			assertThat(resultAttachmentGroupName).isEqualTo(ATTACHMENT_GROUP_NAME);
		}

		private XdomeaMessageData doClassify() {
			return fileClassifier.mapIncomingFilesToXdomeaMessageData(incomingFileList);
		}
	}

}
