/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.xta;

import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.util.List;
import java.util.function.Consumer;

import io.grpc.Channel;
import lombok.RequiredArgsConstructor;
import lombok.SneakyThrows;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.context.annotation.Bean;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.bean.override.mockito.MockitoBean;
import org.springframework.test.context.bean.override.mockito.MockitoSpyBean;

import de.ozgcloud.eingang.Application;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.router.ManagableStub;
import de.ozgcloud.eingang.router.VorgangManagerServerResolver;
import de.ozgcloud.eingang.router.VorgangRemoteService;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceStub;
import de.ozgcloud.vorgang.vorgang.GrpcEingang;
import de.ozgcloud.vorgang.vorgang.VorgangServiceGrpc.VorgangServiceBlockingStub;
import de.ozgcloud.xta.client.XtaClient;
import de.ozgcloud.xta.client.model.XtaMessage;

@SpringBootTest(classes = {
		Application.class,
		XtaITCase.ActivateXTARunnerConfig.class
})
@ActiveProfiles({ "itcase" })
class XtaITCase {

	@MockitoSpyBean
	private VorgangRemoteService vorgangRemoteService;
	@MockitoBean
	private VorgangManagerServerResolver resolver;

	@Mock
	private ManagableStub<VorgangServiceBlockingStub> vorgangManageableServiceStub;
	@Mock
	private VorgangServiceBlockingStub vorgangServiceStub;
	@Mock
	private ManagableStub<BinaryFileServiceStub> binaryFileManageableServiceStub;
	@Mock
	private BinaryFileServiceStub binaryFileServiceStub;
	@Mock
	private Channel channel;

	@MockitoBean
	private XtaClient xtaClient;

	@Captor
	private ArgumentCaptor<FormData> formDataArgumentCaptor;
	@Captor
	private ArgumentCaptor<GrpcEingang> eingangArgumentCaptor;

	@Autowired
	private XtaRunner runner;

	@BeforeEach
	void initTest() {
		when(resolver.resolveVorgangServiceBlockingStubByOrganisationseinheitenId(any())).thenReturn(vorgangManageableServiceStub);
		when(vorgangManageableServiceStub.get()).thenReturn(vorgangServiceStub);
		when(vorgangServiceStub.getChannel()).thenReturn(channel);
		when(channel.authority()).thenReturn("authorityDummy");

		when(resolver.resolveBinaryFileServiceStubByOrganisationsEinheitId(any())).thenReturn(binaryFileManageableServiceStub);
		when(binaryFileManageableServiceStub.get()).thenReturn(binaryFileServiceStub);
	}

	@DisplayName("run get xta messages")
	@Nested
	class TestRunGetXtaMessages {
		@BeforeEach
		void setup() {
			doReturn("vorgangId(unused)").when(vorgangRemoteService).createVorgang(any(), any(), any(), any());
		}

		@DisplayName("should have three representations with pdf")
		@TestZipFileNamesWithPdf
		void shouldHaveThreeRepresentationsWithPdf(String zipFileName) {
			mockNachrichtenBroker(zipFileName);

			runner.runGetXtaMessages();

			var eingang = captureEingang();
			assertThat(eingang.getNumberOfRepresentations()).isEqualTo(3);
		}

		@DisplayName("should send two representations without pdf")
		@TestZipFileNamesWithoutPdf
		void shouldSendTwoXMLRepresentationsWithoutPdf(String zipFileName) {
			mockNachrichtenBroker(zipFileName);

			runner.runGetXtaMessages();

			var eingang = captureEingang();
			assertThat(eingang.getNumberOfRepresentations()).isEqualTo(2);
		}

		@DisplayName("should have no attachments")
		@TestZipFileNamesWithoutAnlage
		void shouldHaveNoAttachments(String zipFileName) {
			mockNachrichtenBroker(zipFileName);

			runner.runGetXtaMessages();

			var eingang = captureEingang();
			assertThat(eingang.getNumberOfAttachments()).isZero();
		}

		@DisplayName("should have one attachment")
		@TestZipFileNamesWithAnlage
		void shouldHaveOneAttachment(String zipFileName) {
			mockNachrichtenBroker(zipFileName);

			runner.runGetXtaMessages();

			var eingang = captureEingang();
			assertThat(eingang.getNumberOfAttachments()).isEqualTo(1);
		}

		@DisplayName("should have postfach id")
		@TestZipFileNamesWithServiceKonto
		void shouldHavePostfachId(String zipFileName) {
			mockNachrichtenBroker(zipFileName);

			runner.runGetXtaMessages();

			var serviceKonto = captureEingang()
					.getHeader()
					.getServiceKonto();
			assertThat(serviceKonto).isNotNull();
			var postfachId = serviceKonto
					.getPostfachAddresses(0)
					.getIdentifier()
					.getProperty(0)
					.getValue(0);
			assertThat(postfachId).isEqualTo("4dd01647-b9d9-4775-1b50-08da3d83800a");
		}

		@DisplayName("should have organistationseinheit ID")
		@TestZipFileNames
		void shouldHaveOrganistationseinheitId(String zipFileName) {
			mockNachrichtenBroker(zipFileName);

			runner.runGetXtaMessages();

			var organisationseinheitenId = captureEingang()
					.getZustaendigeStelle()
					.getOrganisationseinheitenId();
			assertThat(organisationseinheitenId).isEqualTo("9795669");
		}

		@DisplayName("should have generated vorgang nummer for dFoerdermittel")
		@Test
		void shouldHaveVorgangsnummer() {
			mockNachrichtenBroker("dfoerdermittel_with_anlage.zip");

			runner.runGetXtaMessages();

			var vorgangsNummer = captureEingang()
					.getHeader()
					.getVorgangNummer();
			assertThat(vorgangsNummer).hasSameSizeAs("4627-TY3X");
		}

		@DisplayName("should have vorgang nummer for AFM")
		@TestZipFileNamesFromAFM
		void shouldHaveVorgangsnummerForAfm(String zipFileName) {
			mockNachrichtenBroker(zipFileName);

			runner.runGetXtaMessages();

			var vorgangsNummer = captureEingang()
					.getHeader()
					.getVorgangNummer();
			assertThat(vorgangsNummer).hasSameSizeAs("20240404370530710707");
		}

		@DisplayName("should use xta vorgangsnummer for FIM")
		@Test
		void shouldUseVorgangsnummerForFIM() {
			mockNachrichtenBroker("versammlungsanzeige.xml");

			runner.runGetXtaMessages();

			var vorgangNummer = captureEingang().getHeader().getVorgangNummer();
			assertThat(vorgangNummer).hasSize(9);
		}

		@DisplayName("should have one attachment for FIM")
		@Test
		void shouldHaveOneRepresentationForFim() {
			mockNachrichtenBroker("versammlungsanzeige.xml");

			runner.runGetXtaMessages();

			var eingang = captureEingang();
			assertThat(eingang.getNumberOfRepresentations()).isEqualTo(1);
		}

		private GrpcEingang captureEingang() {
			verify(vorgangRemoteService, times(1))
					.createVorgang(
							formDataArgumentCaptor.capture(),
							eingangArgumentCaptor.capture(), any(), any());

			assertThat(formDataArgumentCaptor.getValue()).isNotNull();

			return eingangArgumentCaptor.getValue();
		}
	}

	@SneakyThrows
	private void mockNachrichtenBroker(String zipFileName) {
		XtaMessage xtaMessage = XtaResponseTestFactory.createGetMessageResponse(zipFileName);
		when(xtaClient.fetchMessages(any())).thenAnswer(invocation -> {
			@SuppressWarnings("unchecked")
			var processMessage = (Consumer<XtaMessage>) invocation.getArgument(0);
			processMessage.accept(xtaMessage);
			return List.of(XtaTransportReportTestFactory.create());
		});
	}

	@Target({ ElementType.METHOD })
	@Retention(RetentionPolicy.RUNTIME)
	@ParameterizedTest
	@ValueSource(strings = {
			"dfoerdermittel_without_anlage.zip",
			"brauchtumsfeuer_without_anlage.zip",
			"mantelantrag_without_anlage.zip"
	})
	@interface TestZipFileNamesWithoutAnlage {
	}

	@Target({ ElementType.METHOD })
	@Retention(RetentionPolicy.RUNTIME)
	@ParameterizedTest
	@ValueSource(strings = {
			"dfoerdermittel_with_anlage.zip",
			"brauchtumsfeuer_with_anlage.zip",
			"mantelantrag_with_anlage.zip"
	})
	@interface TestZipFileNamesWithAnlage {
	}

	@Target({ ElementType.METHOD })
	@Retention(RetentionPolicy.RUNTIME)
	@ParameterizedTest
	@ValueSource(strings = {
			"dfoerdermittel_without_anlage.zip",
			"mantelantrag_without_anlage.zip",
			"brauchtumsfeuer_without_anlage.zip",
			"dfoerdermittel_with_anlage.zip",
			"mantelantrag_with_anlage.zip",
			"brauchtumsfeuer_with_anlage.zip",
	})
	@interface TestZipFileNamesWithPdf {
	}

	@Target({ ElementType.METHOD })
	@Retention(RetentionPolicy.RUNTIME)
	@ParameterizedTest
	@ValueSource(strings = {
			"waffenschein.zip"
	})
	@interface TestZipFileNamesWithoutPdf {
	}

	@Target({ ElementType.METHOD })
	@Retention(RetentionPolicy.RUNTIME)
	@ParameterizedTest
	@ValueSource(strings = {
			"mantelantrag_without_anlage.zip",
			"brauchtumsfeuer_without_anlage.zip",
			"mantelantrag_with_anlage.zip",
			"brauchtumsfeuer_with_anlage.zip",
	})
	@interface TestZipFileNamesFromAFM {
	}

	@Target({ ElementType.METHOD })
	@Retention(RetentionPolicy.RUNTIME)
	@ParameterizedTest
	@ValueSource(strings = {
			"dfoerdermittel_without_anlage.zip",
			"dfoerdermittel_with_anlage.zip",
	})
	@interface TestZipFileNamesWithServiceKonto {
	}

	@Target({ ElementType.METHOD })
	@Retention(RetentionPolicy.RUNTIME)
	@ParameterizedTest
	@ValueSource(strings = {
			"dfoerdermittel_without_anlage.zip",
			"mantelantrag_without_anlage.zip",
			"brauchtumsfeuer_without_anlage.zip",
			"dfoerdermittel_with_anlage.zip",
			"mantelantrag_with_anlage.zip",
			"brauchtumsfeuer_with_anlage.zip",
			"waffenschein.zip"
	})
	@interface TestZipFileNames {
	}

	@RequiredArgsConstructor
	static class ActivateXTARunnerConfig {
		private final XtaService xtaService;

		@Bean
		XtaRunner xtaRunner() {
			return new XtaRunner(xtaService);
		}
	}
}
