/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.xta;

import static de.ozgcloud.eingang.xta.zip.TestZipFileFactory.*;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.nio.charset.StandardCharsets;
import java.time.ZonedDateTime;
import java.util.Arrays;
import java.util.Collections;
import java.util.Map;
import java.util.Objects;
import java.util.UUID;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import jakarta.activation.DataHandler;
import jakarta.activation.DataSource;
import jakarta.activation.FileDataSource;
import org.apache.commons.codec.Resources;
import org.apache.commons.io.IOUtils;

import de.ozgcloud.common.binaryfile.TempFileUtils;
import de.ozgcloud.xta.client.model.XtaFile;
import de.ozgcloud.xta.client.model.XtaMessage;
import de.ozgcloud.xta.client.model.XtaMessageMetaData;

public class XtaResponseTestFactory {

	public static final Map<String, String> MESSAGE_TYPE_BY_ATTACHMENT_FILENAME = Map.of(
			"mantelantrag_without_anlage.zip", "Geschaeftsgang.Geschaeftsgang.0201",
			"dfoerdermittel_without_anlage.zip", "Geschaeftsgang.Geschaeftsgang.0201",
			"brauchtumsfeuer_without_anlage.zip", "Geschaeftsgang.Geschaeftsgang.0201",
			"versammlungsanzeige.xml", "fim.S17000652.17000652001004",
			"mantelantrag_with_anlage.zip", "Geschaeftsgang.Geschaeftsgang.0201",
			"dfoerdermittel_with_anlage.zip", "Geschaeftsgang.Geschaeftsgang.0201",
			"brauchtumsfeuer_with_anlage.zip", "Geschaeftsgang.Geschaeftsgang.0201",
			"waffenschein.zip", "Geschaeftsgang.Geschaeftsgang.0201"
	);

	private static final Map<String, String> MESSAGE_ID_BY_ATTACHMENT_FILENAME = MESSAGE_TYPE_BY_ATTACHMENT_FILENAME
			.keySet().stream()
			.collect(Collectors.toMap(name -> name, name -> generateMessageID()));

	private static String generateMessageID() {
		return "urn:de:xta:messageid:dataport_xta_210:%s".formatted(UUID.randomUUID().toString());
	}

	public static XtaMessage createGetMessageResponse(String xtaAttachmentFileName) {
		File file;
		try (var inputStream = getAttachmentInputStream(xtaAttachmentFileName)) {
			file = TempFileUtils.writeTmpFile(inputStream);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}

		DataSource dataSource = new FileDataSource(file);
		DataHandler dataHandler = new DataHandler(dataSource);

		return XtaMessage.builder()
				.metaData(createMessageMetaData(xtaAttachmentFileName))
				.messageFile(XtaFile.builder()
						.contentType("application/zip")
						.name(xtaAttachmentFileName)
						.size(BigInteger.valueOf(file.length()))
						.content(dataHandler)
						.build())
				.attachmentFiles(Collections.emptyList())
				.build();
	}

	private static XtaMessageMetaData createMessageMetaData(String xtaAttachmentFileName) {
		return XtaMessageMetaDataTestFactory.createBuilder()
				.messageId(new XtaMessageId(MESSAGE_ID_BY_ATTACHMENT_FILENAME.get(xtaAttachmentFileName)).toString())
				.messageTypeCode(getMessageType(xtaAttachmentFileName))
				.deliveryAttributesOrigin(ZonedDateTime.now())
				.deliveryAttributesDelivery(ZonedDateTime.now())
				.build();
	}

	private static InputStream getAttachmentInputStream(String xtaAttachmentFileName) {
		var attachmentResourcePath = getAttachmentFilePath(xtaAttachmentFileName);
		if (xtaAttachmentFileName.endsWith(".zip")) {
			try {
				return new FileInputStream(
						createAttachmentZipFile(
								attachmentResourcePath.replace(".zip", "")));
			} catch (FileNotFoundException e) {
				throw new RuntimeException(e);
			}
		} else {
			return Resources.getInputStream(attachmentResourcePath);
		}
	}

	private static File createAttachmentZipFile(String attachmentResourceDirectory) {
		return createTempZipFile(listResourceFileNames(attachmentResourceDirectory)
				.map(subName -> {
							try {
								return TestZipEntry.builder()
										.name(subName)
										.content(IOUtils.toByteArray(Resources.getInputStream(attachmentResourceDirectory + "/" + subName)))
										.build();
							} catch (IOException e) {
								throw new RuntimeException(e);
							}
						}
				).toList());
	}

	private static Stream<String> listResourceFileNames(String resourceDirectoryPath) {
		try (var inputStream = Resources.getInputStream(resourceDirectoryPath)) {
			return Arrays.stream(IOUtils.toString(inputStream, StandardCharsets.UTF_8).split("\n"));
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	private static String getMessageType(String xtaAttachmentFileName) {
		return Objects.requireNonNull(
				MESSAGE_TYPE_BY_ATTACHMENT_FILENAME.get(xtaAttachmentFileName),
				"Xta-message type for '%s' has to be configured!".formatted(xtaAttachmentFileName)
		);
	}

	private static String getAttachmentFilePath(String xtaAttachmentFileName) {
		return "/mock-responses/getMessage/%s/%s".formatted(getMessageType(xtaAttachmentFileName), xtaAttachmentFileName);
	}
}
