/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.xta;

import static de.ozgcloud.eingang.xta.XtaService.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.Optional;

import lombok.SneakyThrows;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormData.Representations;
import de.ozgcloud.eingang.common.formdata.FormDataControlTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;
import de.ozgcloud.eingang.common.vorgang.VorgangNummerSupplier;
import de.ozgcloud.eingang.semantik.SemantikAdapter;
import de.ozgcloud.eingang.xdomea.XdomeaMessageDataMapper;
import de.ozgcloud.xta.client.XtaClient;
import de.ozgcloud.xta.client.model.XtaMessage;
import de.ozgcloud.xta.client.model.XtaMessageMetaData;
import de.ozgcloud.xta.client.model.XtaTransportReport;

class XtaServiceTest {

	@Spy
	@InjectMocks
	private XtaService service;

	@Mock
	private XtaClient xtaClient;
	@Mock
	private XtaMessageMapper mapper;
	@Mock
	private VorgangNummerSupplier vorgangNummerSupplier;
	@Mock
	private SemantikAdapter semantikAdapter;

	@Mock
	private XtaIncomingFilesMapper incomingFilesMapper;

	@Mock
	private XdomeaMessageDataMapper xdomeaMessageDataMapper;

	@Mock
	private FormData formData;

	@Nested
	class TestFetchMessages {

		private final XtaTransportReport transportReport = XtaTransportReportTestFactory.create();

		@BeforeEach
		@SneakyThrows
		void setup() {
			doReturn(List.of(transportReport)).when(xtaClient).fetchMessages(any());
		}

		@Test
		@SneakyThrows
		void shouldUseMessageProcessor() {
			service.fetchMessages();

			verify(xtaClient).fetchMessages(service.messageProcessor);
		}

		@Test
		@SneakyThrows
		void shouldCallLogTransportReports() {
			service.fetchMessages();

			verify(service).logTransportReports(any());
		}
	}

	@DisplayName("process message")
	@Nested
	class TestProcessMessage {
		private final XtaMessage xtaMessage = XtaMessageTestFactory.create();

		@DisplayName("should call getFormDataIfNoRuntimeException")
		@Test
		void shouldCallGetFormDataIfNoRuntimeException() {
			service.processMessage(xtaMessage);

			verify(service).getFormDataIfNoRuntimeException(any());
		}

		@Test
		void shouldCallProcessSemantik() {
			doReturn(Optional.of(formData)).when(service).getFormDataIfNoRuntimeException(any());

			service.processMessage(xtaMessage);

			verify(service).processSemantik(formData);
		}

		@Test
		void shouldNotCallProcessSemantik() {
			doReturn(Optional.empty()).when(service).getFormDataIfNoRuntimeException(any());

			service.processMessage(xtaMessage);

			verify(service, never()).processSemantik(any());
		}
	}

	@DisplayName("get form data if no runtime exception")
	@Nested
	class TestGetFormDataIfNoRuntimeException {

		XtaMessage message;

		@Mock
		FormData formData;

		@BeforeEach
		void before() {
			message = XtaMessageTestFactory.create();
		}

		@DisplayName("should call get formdata")
		@Test
		void shouldCallGetFormdata() {
			service.getFormDataIfNoRuntimeException(message);

			verify(service).getFormData(message);
		}

		@DisplayName("with exception")
		@Nested
		class TestWithException {
			@DisplayName("should return empty")
			@Test
			void shouldReturnEmpty() {
				doThrow(new RuntimeException("test-error")).when(service).getFormData(any());

				var result = service.getFormDataIfNoRuntimeException(message);

				assertThat(result).isEmpty();
			}
		}

		@DisplayName("without exception")
		@Nested
		class TestWithoutException {
			@DisplayName("should return")
			@Test
			void shouldReturn() {
				doReturn(formData).when(service).getFormData(any());

				var result = service.getFormDataIfNoRuntimeException(message);

				assertThat(result).contains(formData);
			}
		}
	}

	@DisplayName("get form data")
	@Nested
	class TestGetFormData {

		private final XtaMessage message = XtaMessageTestFactory.create();
		private final XtaMessageMetaData messageMetaData = XtaMessageMetaDataTestFactory.create();
		private final FormData mappedFormData = FormDataTestFactory.create();
		private RepresentationsAttachmentsPair classification;

		@Captor
		private ArgumentCaptor<XtaMessageMetaData> messageMetaDataCaptor;

		@Captor
		private ArgumentCaptor<RepresentationsAttachmentsPair> classificationCaptor;

		@BeforeEach
		void init() {
			classification = RepresentationsAttachmentsPairTestFactory.create();
			var incomingFiles = List.of(IncomingFileTestFactory.create(), IncomingFileTestFactory.create());
			when(incomingFilesMapper.toIncomingFiles(message.messageFile())).thenReturn(incomingFiles);
			doReturn(classification).when(service).getRepresentationsAttachmentsPair(messageMetaData, incomingFiles);
		}

		@BeforeEach
		void mockMessageMapping() {
			when(mapper.toFormData(any(), any(), any())).thenReturn(mappedFormData);
		}

		@Test
		void shouldCallMapper() {
			service.getFormData(message);

			verify(mapper).toFormData(classification, messageMetaData, vorgangNummerSupplier);
		}

		@Test
		void shouldCallMapperToFormData() {
			service.getFormData(message);

			verify(mapper).toFormData(classificationCaptor.capture(), messageMetaDataCaptor.capture(), eq(vorgangNummerSupplier));
			assertThat(messageMetaDataCaptor.getValue()).isEqualTo(messageMetaData);
			assertThat(classificationCaptor.getValue()).isEqualTo(classification);
		}

		@Test
		void shouldCallAddAttachments() {
			service.getFormData(message);

			verify(service).addAttachments(message, mappedFormData);
		}

		@Test
		void shouldReturnMappedResult() {
			doReturn(mappedFormData).when(service).addRepresentations(any(), any());

			var result = service.getFormData(message);

			assertThat(result).isEqualTo(mappedFormData);
		}
	}

	@DisplayName("add attachments")
	@Nested
	class TestAddAttachments {

		private final FormData inFormData = FormDataTestFactory.createBuilder().clearAttachments().numberOfAttachments(0).build();

		@Test
		void shouldAddAttachments() {
			var result = service.addAttachments(XtaMessageTestFactory.create(), inFormData);

			assertThat(result.getAttachments()).hasSize(1);
			assertThat(result.getNumberOfAttachments()).isEqualTo(1);
		}

		@Test
		void shouldWorkWithoutAnyAttachment() {
			var result = service.addAttachments(XtaMessageTestFactory.createBuilder().attachmentFiles(Collections.emptyList()).build(), inFormData);

			assertThat(result.getAttachments()).isEmpty();
			assertThat(result.getNumberOfAttachments()).isZero();
		}
	}

	@DisplayName("get representations attachments pair")
	@Nested
	class TestGetRepresentationsAttachmentsPair {
		@Mock
		private XtaMessageMetaData messageMetaData;

		@Mock
		private RepresentationsAttachmentsPair classification;

		private List<IncomingFile> incomingFiles;

		@BeforeEach
		void mock() {
			incomingFiles = List.of(IncomingFileTestFactory.create(), IncomingFileTestFactory.create());
		}

		@DisplayName("should use correct mapper xdomea message type")
		@Test
		void shouldUseCorrectMapperXdomeaMessageType() {
			when(messageMetaData.messageTypeCode()).thenReturn(XtaService.XDOMEA_0201_MESSAGE_TYPE);
			when(xdomeaMessageDataMapper.mapIncomingFilesToRepresentationsAttachmentsPair(incomingFiles))
					.thenReturn(classification);

			var result = service.getRepresentationsAttachmentsPair(messageMetaData, incomingFiles);

			assertThat(result).isEqualTo(classification);
		}

		@DisplayName("should use correct mapping for FIM message")
		@Test
		void shouldUseCorrectMappingForFimMessage() {
			when(messageMetaData.messageTypeCode()).thenReturn(FIM_MESSAGE_TYPE_PREFIX + "836487");
			doReturn(classification).when(service).mapIncomingFilesToRepresentations(incomingFiles);

			var result = service.getRepresentationsAttachmentsPair(messageMetaData, incomingFiles);

			assertThat(result).isEqualTo(classification);
		}

		@DisplayName("should throw exception for unexpected message type")
		@Test
		void shouldThrowExceptionForUnexpectedMessageType() {
			when(messageMetaData.messageTypeCode()).thenReturn("unexpected");

			assertThatThrownBy(() -> service.getRepresentationsAttachmentsPair(messageMetaData, incomingFiles))
					.isInstanceOf(TechnicalException.class);
		}
	}

	@DisplayName("map incoming files to representations")
	@Nested
	class TestMapIncomingFilesToRepresentations {

		private List<IncomingFile> incomingFiles;

		@BeforeEach
		void mock() {
			incomingFiles = List.of(IncomingFileTestFactory.create(), IncomingFileTestFactory.create());

		}

		@DisplayName("should return representations with incoming files")
		@Test
		void shouldReturnRepresentationsWithIncomingFiles() {
			var result = service.mapIncomingFilesToRepresentations(incomingFiles);

			assertThat(result.representations()).isEqualTo(incomingFiles);
		}

		@DisplayName("should return attachments with empty list")
		@Test
		void shouldReturnAttachmentsWithEmptyList() {
			var result = service.mapIncomingFilesToRepresentations(incomingFiles);

			assertThat(result.attachments()).isEmpty();
		}
	}

	@DisplayName("add representations")
	@Nested
	class TestAddRepresentations {
		@Test
		void shouldAddPrimaryRepresentation() {
			var result = service.addRepresentations(FormDataTestFactory.create(), XtaMessageTestFactory.PRIMARY_FORM_DATA_MESSAGE);

			assertThat(result.getControl().getRepresentations()).isPresent().get()
					.extracting(Representations::getPrimaryFormDataRepresentation).isEqualTo(XtaMessageTestFactory.PRIMARY_FORM_DATA_MESSAGE);
		}

		@Test
		void shouldRespectExistingRepresentation() {
			var formDataWithRepresentation = FormDataTestFactory.createBuilder().control(FormDataControlTestFactory.createBuilder()
					.representations(Optional.of(
							Representations.builder().primaryFormDataPdfRepresentation("PDF_FILE").build()))
					.build()).build();

			var result = service.addRepresentations(formDataWithRepresentation, XtaMessageTestFactory.PRIMARY_FORM_DATA_MESSAGE);

			var baseAssert = assertThat(result.getControl().getRepresentations()).isPresent().get();
			baseAssert.extracting(Representations::getPrimaryFormDataRepresentation).isEqualTo(XtaMessageTestFactory.PRIMARY_FORM_DATA_MESSAGE);
			baseAssert.extracting(Representations::getPrimaryFormDataPdfRepresentation).isEqualTo("PDF_FILE");
		}
	}

	@DisplayName("process semantik")
	@Nested
	class TestProcessSemantik {
		@Test
		void shouldCallSemantikAdapter() {
			service.processSemantik(FormDataTestFactory.create());

			verify(semantikAdapter).processFormData(any());
		}
	}

}
