/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.xta.zip;

import static de.ozgcloud.eingang.xta.zip.LimitedInputStream.*;
import static org.assertj.core.api.Assertions.*;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;

import org.apache.commons.io.IOUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;

import lombok.SneakyThrows;

class LimitedInputStreamTest {

	private static final int READ_LIMIT = 10;
	private static final String STRING_WITH_READ_LIMIT_LENGTH = "A".repeat(READ_LIMIT);
	private static final String STRING_WITH_MORE_THAN_READ_LIMIT_LENGTH = "B".repeat(READ_LIMIT + 1);

	private LimitedInputStream limitedInputStream;

	private InputStream createStringInputSteam(String string) {
		return new ByteArrayInputStream(string.getBytes());
	}

	@SneakyThrows
	private String readInputStreamToString(InputStream inputStream) {
		return IOUtils.toString(inputStream, Charset.defaultCharset());
	}

	@SneakyThrows
	@DisplayName("should succeed if read limit is not exceeded")
	@Test
	void shouldSucceedIfReadLimitIsNotExceeded() {
		limitedInputStream = new LimitedInputStream(createStringInputSteam(STRING_WITH_READ_LIMIT_LENGTH), READ_LIMIT);

		var outputString = readInputStreamToString(limitedInputStream);

		assertThat(outputString).isEqualTo(STRING_WITH_READ_LIMIT_LENGTH);
	}

	@DisplayName("should fail if read limit is exceeded")
	@Test
	void shouldFailIfReadLimitIsExceeded() {
		limitedInputStream = new LimitedInputStream(createStringInputSteam(STRING_WITH_MORE_THAN_READ_LIMIT_LENGTH), READ_LIMIT);

		assertThatThrownBy(() -> readInputStreamToString(limitedInputStream))
				.isInstanceOf(IOException.class)
				.hasMessage(LIMITED_EXCEEDED_MESSAGE);
	}

	@DisplayName("read")
	@Nested
	class TestRead {

		@SneakyThrows
		@DisplayName("should return")
		@Test
		void shouldReturn() {
			limitedInputStream = createLimitedInputStream();

			var result = limitedInputStream.read();
			assertThat(result).isEqualTo(STRING_WITH_READ_LIMIT_LENGTH.getBytes()[0]);
		}

		@SneakyThrows
		@DisplayName("should advance bytesRead")
		@Test
		void shouldAdvanceBytesRead() {
			limitedInputStream = createLimitedInputStream();
			limitedInputStream.bytesRead = READ_LIMIT - 1;

			limitedInputStream.read();
			assertThat(limitedInputStream.bytesRead).isEqualTo(READ_LIMIT);
		}

		@DisplayName("should throw if exceeded")
		@Test
		void shouldThrowIfExceeded() {
			limitedInputStream = createLimitedInputStreamWithExceeding();
			limitedInputStream.bytesRead = READ_LIMIT;

			assertThatThrownBy(() -> limitedInputStream.read()).isInstanceOf(IOException.class);
		}

	}

	@DisplayName("read into buffer")
	@Nested
	class TestReadIntoBuffer {
		private byte[] buffer;

		@BeforeEach
		void mock() {
			buffer = new byte[READ_LIMIT];
		}

		@SneakyThrows
		@DisplayName("should return")
		@Test
		void shouldReturn() {
			limitedInputStream = createLimitedInputStream();

			var result = limitedInputStream.read(buffer);

			assertThat(result).isEqualTo(READ_LIMIT);
		}

		@SneakyThrows
		@DisplayName("should advance bytesRead")
		@Test
		void shouldAdvanceBytesRead() {
			limitedInputStream = createLimitedInputStream();

			limitedInputStream.read(buffer);
			assertThat(limitedInputStream.bytesRead).isEqualTo(READ_LIMIT);
		}

		@DisplayName("should throw if exceeded")
		@Test
		void shouldThrowIfExceeded() {
			limitedInputStream = createLimitedInputStreamWithExceeding();
			limitedInputStream.bytesRead = 1;

			assertThatThrownBy(() -> limitedInputStream.read(buffer)).isInstanceOf(IOException.class);
		}
	}

	private LimitedInputStream createLimitedInputStream() {
		return new LimitedInputStream(createStringInputSteam(STRING_WITH_READ_LIMIT_LENGTH), READ_LIMIT);
	}

	private LimitedInputStream createLimitedInputStreamWithExceeding() {
		return new LimitedInputStream(createStringInputSteam(STRING_WITH_MORE_THAN_READ_LIMIT_LENGTH), READ_LIMIT);
	}
}
