/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.xta.zip;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import org.apache.commons.io.IOUtils;

import de.ozgcloud.common.binaryfile.TempFileUtils;
import lombok.Builder;
import lombok.Getter;

public class TestZipFileFactory {

	private static final String EXPANDED_ENTRY_NAME = "bomb.txt";

	@Builder
	@Getter
	public static class TestZipEntry {
		private String name;
		private byte[] content;
	}

	public static File createTempZipFile(List<TestZipEntry> testZipEntries) {
		var file = TempFileUtils.createTmpFile().toFile();
		try (var zipOutputStream = new ZipOutputStream(new FileOutputStream(file))) {
			for (TestZipEntry entry : testZipEntries) {
				zipOutputStream.putNextEntry(new ZipEntry(entry.getName()));
				zipOutputStream.write(entry.getContent());
				zipOutputStream.closeEntry();
			}
			return file;
		} catch (IOException e) {
			throw new RuntimeException("Failed to create temporary zip file", e);
		}
	}

	public static File createTempZipBomb(int maxTotalSize) {
		return overwriteFileWithZipEntrySize(
				createTempZipWithSingleEntry(maxTotalSize * 2),
				maxTotalSize
		);
	}

	private static File createTempZipWithSingleEntry(int entrySize) {
		var file = TempFileUtils.createTmpFile().toFile();
		try (var zipOutputStream = new ZipOutputStream(new FileOutputStream(file))) {
			var entry = new ZipEntry(EXPANDED_ENTRY_NAME);
			var content = "A".repeat(entrySize).getBytes();

			zipOutputStream.putNextEntry(entry);
			zipOutputStream.write(content);
			zipOutputStream.closeEntry();

		} catch (IOException e) {
			throw new RuntimeException("Failed to create temporary zip file", e);
		}
		return file;
	}

	private static File overwriteFileWithZipEntrySize(File file, int newSize) {
		try {
			var zipFileBytes = IOUtils.toByteArray(new FileInputStream(file));
			overwriteZipEntrySize(zipFileBytes, EXPANDED_ENTRY_NAME, newSize);

			// Write the adjusted ZIP content back to the file
			try (var fos = new FileOutputStream(file)) {
				fos.write(zipFileBytes);
			}
		} catch (IOException e) {
			throw new RuntimeException("Failed to adjust size header of zip file", e);
		}
		return file;
	}

	private static void overwriteZipEntrySize(byte[] zipFileBytes, String entryName, int newSize) throws IOException {
		// Modify the uncompressed size entry size in the central directory structure (which is located at the end)
		// Zip structure spec: https://www.iana.org/assignments/media-types/application/zip
		var entryNameBytes = entryName.getBytes();

		var lastIndexOfEntryName = findLastStartIndex(zipFileBytes, entryNameBytes);
		if (lastIndexOfEntryName == -1) {
			throw new IOException("ZIP entry not found: " + entryName);
		}
		var uncompressedSizeFieldStartOffset = lastIndexOfEntryName - (4 * 2 + 5 * 2 + 4);
		writeIntToByteArray(newSize, zipFileBytes, uncompressedSizeFieldStartOffset);
	}

	private static void writeIntToByteArray(int value, byte[] array, int offset) {
		array[offset] = (byte) (value & 0xFF);
		array[offset + 1] = (byte) ((value >> 8) & 0xFF);
		array[offset + 2] = (byte) ((value >> 16) & 0xFF);
		array[offset + 3] = (byte) ((value >> 24) & 0xFF);
	}

	private static int findLastStartIndex(byte[] haystack, byte[] needle) {
		var matchOffset = 0;
		for (var i = haystack.length - 1; i >= needle.length; i--) {
			if (haystack[i] == needle[needle.length - 1 - matchOffset]) {
				if (++matchOffset == needle.length) {
					return i;
				}
			} else {
				matchOffset = 0;
			}
		}
		return -1;
	}
}
